# Copyright (C) 2019 demetriou engineering ltd.
# Part of `dx-utilities`, licensed under the AGPLv3+
from dx_utilities.units import Mass, Weight


class BaseMaterial(object):
    """Class to represent materials."""

    #: Poisson's ratio
    nu = None
    #: ``[N/m/m/m]``
    weight_density = None
    #: Safety factor
    gamma = None

    def __init__(self):
        self._m = None

    @property
    def w(self):
        """Alias for weight-density ``[N/m/m/m]``.

        :rtype: Weight
        """
        return self.weight_density

    @property
    def m(self):
        """Mass density ``[kg/m/m/m]``.

        :rtype: Mass
        """
        if self._m is None:
            self._m = self.w.to_mass()
        return self._m

    def safety_factor(self, design_situation='persistent',
                      limit_state='ultimate'):
        """Wrapper around the safety factor of the material.
        Returns the value depending on the design situation
        and the limit state considered.

        :param str design_situation:
        :param str limit_state:
        :rtype: float
        """
        return getattr(getattr(self.gamma, design_situation), limit_state)


class BaseReinforcementSteel(BaseMaterial):

    #: Poisson's ratio
    nu = 0.3
    #: ``[N/m/m/m]``
    weight_density = 77078

    def __init__(self):
        self._E = 200 # Modulus of elasticity [GPa]

    @property
    def E(self):
        """Modulus of elasticity ``[N/m/m]``."""
        return self._E*1e+09

    @property
    def G(self):
        """Shear modulus ``[N/m/m]``."""
        return self.E / 2 / (1 + self.nu)



# Copyright (C) 2019 demetriou engineering ltd.
# Part of `dx-utilities`, licensed under the AGPLv3+
"""Collection of abstractions representing forces
and stresses.
"""
from ctypes import c_double

import numpy as np

from dx_utilities.array_interface import typestr


class NodalForces(object):
    """Set of forces and moments on a node, with 6
    degrees-of-freedom in the most general case.

    All values refer to a local-coordinate system to
    be defined on top.

    :param str LC: Name of the load-case or combination.
    :param float Rx: Force in the x direction
    :param float Ry: Force in the y direction
    :param float Rz: Force in the z direction
    :param float Mx: Moment in the x direction
    :param float My: Moment in the y direction
    :param float Mz: Moment in the z direction
    """

    #: Degrees of freedom
    dof = 6

    def __init__(self, LC="LC0", Rx=0., Ry=0., Rz=0., Mx=0., My=0., Mz=0.):
        self.LC = LC
        self.Rx = Rx
        self.Ry = Ry
        self.Rz = Rz
        self.Mx = Mx
        self.My = My
        self.Mz = Mz

    @property
    def container(self):
        """List of nodal forces ``[Rx, Ry, Rz, Mx, My, Mz]``

        :rtype: list
        """
        return [self.Rx, self.Ry, self.Rz, self.Mx, self.My, self.Mz]

    def __iter__(self):
        self.i = 0
        return self

    def __next__(self):
        if self.i == self.dof:
            raise StopIteration
        value = self.container[self.i]
        self.i += 1
        return value

    @property
    def typestr(self):
        """A string providing the basic type of the homogeneous array.
        For more details see the docs on `array interface`_.

        :rtype: str
        """
        return typestr()

    @property
    def ctypes(self):
        """c-type representation of the container.

        :rtype: array of c-double
        """
        array_type = c_double * self.dof
        data = array_type()

        i = 0
        for v in self:
            data[i] = v
            i += 1

        return data

    def array_interface(self):
        """Construct the `numpy` `array interface`_ for class instances.
        The array interface enables the vectorization of the class
        through `numpy.array`.

        :rtype: dict
        """
        ai = {'shape': (1, self.dof), 'typestr': self.typestr, 'version': 3}
        ai['data'] = self.ctypes
        return ai

    __array_interface__ = property(array_interface)

    def __add__(self, other):
        new_container = np.array(self.container) + np.array(other.container)
        return self.__class__(self.LC, *new_container)


class NMyMz(NodalForces):

    dof = 3

    def __init__(self, LC='LC0', N=0., My=0., Mz=0., **kwargs):
        super().__init__(LC=LC, Rx=N, My=My, Mz=Mz, **kwargs)
        self.N = self.Rx

    @property
    def container(self):
        return [self.N, self.My, self.Mz]



# Copyright (C) 2019 demetriou engineering ltd.
# Part of `dx-utilities`, licensed under the AGPLv3+
from dx_utilities.exceptions import CodedException


__all__ = ['ReinforcementError', 'ImplementationError',
           'UnrecognizedMaterialError']


class ReinforcementError(CodedException):
    pass


class ImplementationError(CodedException):
    pass


class UnrecognizedMaterialError(CodedException):
    """Exception to be raised if a user
    tries to configure a material not supported
    in the dx ecosystem.
    """
    pass



# Copyright (C) 2019 demetriou engineering ltd.
# Part of `dx-utilities`, licensed under the AGPLv3+
"""Collection of abstractions representing structural elements
through a planar simplification of their geometry.
"""
import math
import numpy as np

from collections import defaultdict
from mathutils import Vector

from dx_utilities.geometry import PlanarShape
from dx_utilities.fields import (PointValue, UniformPlanarField,
                                 DiscretePlanarField)
from dx_utilities.vectors import AXIS2IDX
from dx_utilities.decorators import tolerate_input
from dx_utilities.exceptions import CodedValueError

from ..exceptions import ReinforcementError


class SteelReinforcement(object):
    """Represent amount of steel reinforcement.

    :param float As: The total area of the reinforcement.
    :param float effective_depth: The effective_depth of
        the bars on the section of interest.
    :param float spacing: The spacing between the bars.
    :param int nbars: The number of bars used.
    """

    #: Nominal bar diameters (EN 10080)
    dnominal = list(range(6, 23, 2))
    dnominal.extend([25, 28, 32, 40])
    dnominal = {dnom*1e-03 for dnom in dnominal}

    def __init__(self, As, effective_depth, spacing=1., nbars=1):
        self.d = effective_depth
        self.s = spacing
        self.total_area = As
        self.area_per_unit_breadth = self.total_area / self.s
        self.nbars = nbars
        self._phi = None

    @property
    def phi(self):
        """The diameter of the bars. If not explicitly given in
        instantiation, it is evaluated based on the total
        reinforcment area ``As`` and the number of bars ``nbars``.

        :rtype: float
        """
        if self._phi is None:
            phi = 2 * math.sqrt(self.As/self.nbars/math.pi)
            if self._phi not in self.dnominal:
                for dnom in self.dnominal:
                    if math.isclose(phi, dnom, rel_tol=1e-03):
                        self._phi = dnom
                        break
                if self._phi is None:
                    self._phi = min(filter(lambda d: d > phi, self.dnominal))
        return self._phi

    @classmethod
    def from_dimensions(cls, diameter, effective_depth, nbars=1, spacing=1.):
        """Create an instance from raw dimensions.

        :param float diameter:
        :param float effective depth:
        :param int nbars: The number of the bars
        :param float spacing: Spacing between bars
        :rtype: SteelReinforcement
        """
        unit_area = math.pi*diameter**2 / 4
        total_area = nbars * unit_area
        return cls(total_area, effective_depth, spacing, nbars=nbars)

    @property
    def As(self):
        """Alias for total reinforcement area.

        :rtype: float
        """
        return self.total_area

    @As.setter
    def As(self, value):
        self.total_area = value

    @property
    def Asbr(self):
        """Alias for area per unit breadth.

        :rtype: float
        """
        return self.area_per_unit_breadth

    @Asbr.setter
    def Asbr(self, value):
        self.area_per_unit_breadth = value

    def __add__(self, other):
        """Define addition for instances of steel-reinforcement
        as follows:

        * Areas are accumulated.
        * Effective depth and spacing are averaged by weighting
          the associated area per unit breadth.

        If ``other`` is not an instance of this class
        the addition is an identity operation.

        :rtype: SteelReinforcement
        """
        if isinstance(other, self.__class__):
            Asbr = self.Asbr + other.Asbr
            d = (self.Asbr*self.d + other.Asbr*other.d)/Asbr
            s = (self.Asbr*self.s + other.Asbr*other.s)/Asbr
            As = Asbr * s
            return self.__class__(As, d, s)
        return self

    def __radd__(self, other):
        """The reflected addition follows the
        conventions followed in the normal addition.

        :rtype: SteelReinforcement
        """
        return self + other

    def __repr__(self):
        phi = int(self.phi*1e+03)
        s = int(self.s*1e+03)
        return f"T{phi:d}@{s}"


class BaseElement(object):
    """A base-class for structural elements.

    :param material: A material object.
    :type material: BaseMaterial or None
    """

    def __init__(self, material=None, *args, **kwargs):
        self.material = material
        super().__init__(*args, **kwargs)


class PlanarElement(BaseElement, PlanarShape):
    """A base-class for 2D structural elements.

    Any instance may contain two types of objects:

    1. ``agents``: They represent dependent structural elements
       that apply some action, e.g. "columns" on a "slab".
    2. ``fields``: That represent dependent quantities distributed
       in a discrete or uniform manner in the underlying geometry.

    :param float thickness: The thickness of the element
    """
    def __init__(self, thickness=0., *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._thickness = thickness
        # TODO: If linear indexing proves to be inefficient
        #       while manipulating agents and fields,
        #       construct an R-tree so that they are indexed
        #       by their geometry.
        self.agents = defaultdict(list)
        self.fields = defaultdict(list)

    @property
    def thickness(self):
        """Thickness of the element."""
        return self._thickness

    @thickness.setter
    def thickness(self, value):
        self._thickness = value

    def generate_agent(self, shape='rectangle', agent_type=None,
                       *args, **kwargs):
        """Create an agent instance by specifying its shape with
        reference to the coordinate system of the slab. The shape of the agent
        must lie within the geometry of the slab.

        :param str shape: The shape of the agent on the plane.
            Values supported are: ``'rectangle'``, ``'circ'`` or ``'generic'``.
        :param agent_type: The class to be used for the instantiation
            of the agent. Must be a subclass of ``PlanarShape``.
        :raises ValueError: If specified ``shape`` is not supported, or
            if the agent's geometry extends beyond the bounds of the
            slab geometry.
        :param \*args: See positional arguments in the agent-factory
            methods.
        :param \*\*kwargs: See keyword-arguments in the agent-factory
            methods.
        :return: An instance of the ``agent_type`` class.
        """
        if agent_type is None:
            agent_type = self.__class__

        agent = agent_type.new(shape=shape, *args, **kwargs)

        if self.more.contains(agent):
            return agent
        raise CodedValueError(2000, ("Specified geometry extends "
                                     "beyond the bounds of the slab"))

    def add_agent(self, label='', *args, **kwargs):
        """Add an agent instance in the respective container.

        :param str label: The label of the agent to add, e.g. "columns".
        """
        self.agents[label].append(self.generate_agent(*args, **kwargs))

    def generate_field(self, field_class, *args, **kwargs):
        """Generate a field by instantiating the specified
        ``field_class`` from `dx_utilities.fields.fields`.

        :param field_class: The class name of the field.
        :param \*args: See positional arguments in ``field_class``.
        :param \*\*kwargs: See keyword arguments in ``field_class``.
        """
        return field_class(*args, **kwargs)

    def set_field(self, label='', *args, **kwargs):
        """Add a field instance in the respective container.

        :param str label: The label of the field to add.
        :param \*args: See positional arguments in `generate_field`.
        :param \*\*kwargs: See keyword arguments in `generate_field`.
        """
        self.fields[label].append(self.generate_field(*args, **kwargs))


class RCColumn2D(PlanarElement):
    """A 2D representation of reinforced-concrete columns."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._drop_panel = None

    @property
    def drop_panel(self):
        return self._drop_panel

    def set_drop_panel(self, factory=None, *args, **kwargs):
        """Set the value of the ``drop_panel`` property to
        a `ColumnDropPanel` instance (or any subclass thereof)
        generated through the `ColumnDropPanel.create_as_column_offset`
        method.

        :param factory: The class that generates the drop-panel.
        :param \*args: Positional arguments of the wrapped method.
        :param \*\*kwargs: Keyword arguments of the wrapped method.
               """
        if factory is None:
            factory = ColumnDropPanel
        self._drop_panel = factory.create_as_column_offset(
                self, *args, **kwargs
                )


class ColumnDropPanel(PlanarElement):
    """Drop-panels of slab-columns. Best to instantiate
    with `create_as_column_offset`. In that case,
    geometry is defined with reference to the parent-column.

    :param RCColumn2D column:
    :param \*args: Positional arguments in parent class.
    :param \*\*kwargs: Key-word arguments in parent class.
    """
    def __init__(self, column=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.column = column
        self._offset = None
        self._l1 = None
        self._l2 = None

    @classmethod
    @tolerate_input('lx', 'ly', 'height')
    def create_as_column_offset(cls, column, lx, height, ly=None, *args, **kwargs):
        """Generate the projection of the drop-panel as an
        offset of the parent column.

        :param RCColumn2D column: The parent structural element.
        :param float lx: The distance of the drop-panel from the
            column boundary along the x-direction.
        :param float height: The height of the drop-panel.
        :param ly: The distance of the drop-panel from the
            column boundary along the y-direction.
        :type ly: float or None
        :rtype: ColumnDropPanel
        """
        return cls.create_by_offset(column.envelope, dx=lx, dy=ly, column=column,
                                    thickness=height, *args, **kwargs)

    @property
    def height(self):
        return self.thickness

    @property
    def offset(self):
        """Offset distance from the parent column"""
        if self._offset is None:
            self._offset = self.distance_between_bounds(self.column)
        return self._offset

    @property
    def lx(self):
        return self.offset[0]

    @property
    def ly(self):
        return self.offset[1]


class RCSlab(PlanarElement):
    """2D representation of a reinforced-concrete slab.

    Reinforcement is represented as a field within the geometry
    of the slab.

    For raft slabs, or column-bases a soil-pressure field may
    be defined.

    :param str slab_type: The type of the slab. Valid options are:

        * ``'floor-slab'``: Elevated slabs.
        * ``'raft'``: Raft slabs.
        * ``'base'``: Column base.

    :raises ValueError: If an unrecognized slab-type is given.
    """

    #: slab-types that classify as 'foundation'
    foundation_types = {'raft', 'base'}
    #: All the slab-types supported by the class
    slab_types = foundation_types.union({'floor-slab'})

    def __init__(self, slab_type='floor-slab', *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not slab_type in self.slab_types:
            raise CodedValueError(2001, (f'Not valid slab-type. Choose '
                                         f'one of {self.slab_types}.'))
        self.slab_type = slab_type
        self.foundation = self.slab_type in self.foundation_types
        self._sp_index = None

    @property
    def soil_pressure(self):
        """Distribution of soil pressure under the element.
        Obviously, applies only to footings or raft foundations.

        Essentially wraps ``fields['soil-pressure']``

        :rtype: dict
        """
        return self.fields['soil-pressure']

    @property
    def rebar(self):
        """Represent the tension reinforcement though the
        distribution of the area-per-unit-breadth (As), and
        the static-height of the bars over the geometry
        of the slab.

        Essentially wraps the ``fields['rebar']`` field.

        :return: A dictionary ``{'As': list(field), 'd': list(field)}``
        :rtype: dict
        """
        return self.fields['rebar']

    # For backward compatibility
    t_rebar = rebar

    @tolerate_input('value')
    def add_soil_pressure(self, value, region=None):
        """Set uniform value of soil pressure over the specified
        region of the slab. Applies to footings and raft-foundations.
        By default the value of the soil pressure is assumed to be
        uniform throughout the entire geometry of the slab.

        :param float value:
        :param region:
        :type region: `None` or |PlanarShape|
        """
        if region is None:
            region = self
        self.set_field('soil-pressure', field_class=UniformPlanarField,
                       planarshape=region, value=value)

    def add_discrete_soil_pressure(self, coords, values, lc=None):
        """Define a `discrete planar field
        <dx_utilities.fields.fields.DiscretePlanarField>`
        of soil pressure within the geometry of the slab,
        through pairs of coordinates and values.

        The method populates ``fields['soil-pressure']``.

        :param list(tuple) coords: List of tuples ``(x, y)``.
        :param iterable values: Sequence of scalar values of soil-pressure.
        :param lc: The label of the analysis load-case, if any.
        :type lc: str or None
        """
        new_field = DiscretePlanarField.from_coords_and_values(
            coordinates=coords,
            values=np.maximum(np.zeros(len(coords)), list(values)),
            _id=lc
            )
        self.fields['soil-pressure'].append(new_field)

    @property
    def sp_index(self):
        """Index of soil-pressure field according to load-case label

        :return: A map ``{load-case: field}``.
        :rtype: dict()
        """
        if self._sp_index is None:
            self._sp_index = {f.id: f for f in self.soil_pressure}
        return self._sp_index

    def add_uniform_rebar(self, phi, s, d, nbars=1, axes='x',
                          index=None, replace=False):
        """Wrap the `add_rebar_raw` method to configure
        reinforcement on the entire slab.

        :param float phi: The diameter of the longitudinal bars.
        :param float s: Spacing of the reinforcement bars along the
            perpendicular direction.
        :param float d: Distance of the re-bars from the
            farthest compression fibre of the slab.
        :param int nbars: Number of reinforcement bars.
        :param index: The index of the reinforcement in the container.
            If `None`, a new reinforcement is indexed.
        :type index: int or None
        :param bool replace: If `True` replace reinforcement with
            the specified index. Otherwise, update by adding the new
            values. Applies only if ``index`` is not None.
        """
        self.add_rebar_raw(phi=phi, s=s, d=d, nbars=nbars, axes=axes,
                           geometry=self, index=index, replace=replace)

    def add_partial_rebar(self, phi, s, d, nbars=1, axes='x',
                          index=None, replace=False, shape='rectangle',
                          *args, **kwargs):
        """Wrap the `add_rebar_raw` method to add reinforcement
        on a limited region of the slab. This method invokes
        the |PlanarShape.new| to generate the associated geometry.

        :param float phi: The diameter of the longitudinal bars.
        :param float s: Spacing of the reinforcement bars along the
            perpendicular direction.
        :param float d: Distance of the re-bars from the
            farthest compression fibre of the slab.
        :param int nbars: Number of reinforcement bars.
        :param index: The index of the reinforcement in the container.
            If ``None``, a new reinforcement is indexed.
        :type index: int or None
        :param bool replace: If ``True`` replace reinforcement with
            the specified index. Otherwise, update by adding the new
            values. Applies only if ``index`` is not None.
        :param str shape: Shape type to generate.
        :param \*args: Positional arguments of the |PlanarShape.new|.
        :param \*\*kwargs: Positional arguments of the |PlanarShape.new|.
        :raises ValueError: If generated geometry extends beyond the
            bounds of the slab.
        """
        geometry = PlanarShape.new(shape, *args, **kwargs)
        if not self.contains(geometry.less):
            raise CodedValueError(2002, ("Generated geometry extends "
                                         "beyond the bounds of the slab"))
        self.add_rebar_raw(phi=phi, s=s, d=d, nbars=nbars, axes=axes,
                           geometry=geometry, index=index, replace=replace)

    @tolerate_input('phi', 's', 'd')
    @tolerate_input('nbars', 'index', cast=int)
    def add_rebar_raw(self, phi, s, d, nbars=1, axes='x',
                      geometry=None, index=None, replace=False):
        """Construct the representation of reinforcement bars
        from primitive parameters, i.e. diameter, spacing, and distance
        from the farthest compression fibre (effective depth).

        If no ``index`` is given, a new reinforcement representation is
        appended to the respective sequence in the container
        ``fields['rebar']``. Otherwise, the reinforcement stored at the
        specified ``index`` is updated accordingly.

        :param float phi: The diameter of the longitudinal bars.
        :param float s: Spacing of the reinforcement bars along the
            perpendicular direction.
        :param float d: Distance of the re-bars from the
            farthest compression fibre of the slab (effective_depth).
        :param int nbars: Number of reinforcement bars.
        :param str axes: The axial direction of the reinforcement bar.
            To assign the same reinforcement in more than one directions,
            concatenate the axis labels. E.g. ``'xy'`` sets the
            reinforcement in both the x- and the y-direction.
        :param geometry: The geometrical region on the slab where
            the reinforcement lies. If `None` the region coincides
            with the shape of the slab.
        :type geometry: |PlanarShape| or `None`
        :param index: The index of the reinforcement in the container.
            If `None`, a new reinforcement is indexed.
        :type index: int or None
        :param bool replace: If `True` replace reinforcement with
            the specified index. Otherwise, update by adding the new
            values. Applies only if ``index`` is not `None`.
        """
        rebar = SteelReinforcement.from_dimensions(phi, d, spacing=s,
                                                   nbars=nbars)

        vrebar = np.zeros(3, dtype=np.dtype(SteelReinforcement))
        for axis in axes:
            try:
                vrebar[AXIS2IDX[axis.lower()]] = rebar
            except KeyError:
                # Input error
                continue

        if not any(vrebar):
            raise CodedValueError(2003, "No axis specified")

        if geometry is None:
            geometry = self

        if index is None:
            self.set_field('rebar', field_class=UniformPlanarField,
                           planarshape=geometry, value=vrebar)
        else:
            rebar_field = self.fields['rebar'][index]
            if replace:
                rebar_field.replace_value(vrebar)
            else:
                rebar_field.add_value(vrebar)

    # For backward compatibility
    add_tension_rebar_raw = add_rebar_raw

    def resolve_rebar(self, geometry, contains=True):
        """Resolve the tensile reinforcement that is
        found within ``geometry``.

        :param geometry:
        :type geometry: |PlanarShape|
        :param bool contains: If `True` the method
            checks if the fields contain the ``geometry``
            specified. Otherwise, the check is limited
            to verifying intersection.
        :rtype: `list`\ (|UniformPlanarField|)
        """
        if not self.rebar:
            raise ReinforcementError(2004, ("No tensile-reinforcement "
                                            "defined on the slab"))

        containers = []
        check = 'contains' if contains else 'intersects'
        for field in self.rebar:
            if getattr(field.geometry, check)(geometry):
                containers.append(field)

        if not containers:
            raise ReinforcementError(2005, ("No tensile-reinforcement "
                                            "found in the vicinity of "
                                            "the geometry."))

        containers.sort(key=lambda f: f.geometry.intersection(geometry).area,
                        reverse=True)
        return containers



from setuptools import setup, find_packages

long_description = """
# `dx-base` package

A collection of base packages and modules for structural design software.

## Features

* Structural elements in plane.
* Exceptions with error-code.
* Nodal forces.
* Structural materials.
* Material safety factors.

### Sample usage

```
>>> from dx_base.safety_factors import SafetyFactor
>>> sf = SafetyFactor(1.24, 'persistent', 'serviceability')
>>> sf
1.24
>>> sf.design_situation
'persistent'
>>> sf.limit_stte
'serviceability'
```

## Contribute

Source code lives in https://gitlab.com/d-e/dx-base.

## Public API

See the [documentation pages](https://d-e.gitlab.io/dx-base/).
"""

setup(
    name='dx-base',
    packages=find_packages(exclude=['test*']),
    install_requires=[
        "dx-utilities>=1.0.0,<2.0.0",
        ],
    version='1.0.1',
    author="Konstantinos Demartinos",
    author_email="kostas@d-e.gr",
    maintainer="demetriou engineering ltd.",
    maintainer_email="kostas@d-e.gr",
    url="https://gitlab.com/d-e/dx-base",
    description="Base package for structural design software packages",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=["structural-design", ],
    # https://pypi.org/classifiers/
    classifiers=[
        "Programming Language :: Python :: 3",
        ("License :: OSI Approved :: GNU Affero General Public License "
         "v3 or later (AGPLv3+)"),
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        ],
    license='AGPLv3+'
    )



from dx_base.forces import NodalForces, NMyMz
import numpy as np
import unittest


class TestNodalForces(unittest.TestCase):

    def test_array_interface(self):
        nf = NodalForces()
        expected = np.zeros((1, 6))
        np.testing.assert_equal(np.array(nf), expected)

    def test_addition(self):
        nf1 = NodalForces(LC='nf1', Rx=1., Mx=1., My=1.)
        nf2 = NodalForces(LC='nf2', Rx=3., Mx=2., My=9.)
        new_nf = nf1 + nf2
        self.assertAlmostEqual(new_nf.Rx, 4.)
        self.assertAlmostEqual(new_nf.Mx, 3.)
        self.assertAlmostEqual(new_nf.My, 10.)
        self.assertEqual(new_nf.LC, 'nf1')


class TestNMyMz(unittest.TestCase):

    def test_lc(self):
        nf = NMyMz()
        self.assertEqual(nf.LC, 'LC0')

    def test_array_interface(self):
        nf = NMyMz()
        expected = np.zeros((1, 3))
        np.testing.assert_equal(np.array(nf), expected)



import numbers
import unittest

from math import pi
from mathutils import Vector

from dx_base.elements import (RCSlab, RCColumn2D, ColumnDropPanel,
                              SteelReinforcement, PlanarElement)


class TestSteelReinforcement(unittest.TestCase):

    def test_area_per_unit_breadth(self):
        phi = 30.
        d = 45.
        s = 100.
        asx = SteelReinforcement.from_dimensions(phi, d, spacing=s)
        self.assertAlmostEqual(pi*phi**2/4/s, asx.Asbr)

    def test_addition(self):
        d1 = 45.
        phi1 = 30.
        s1 = 100.
        as1 = SteelReinforcement.from_dimensions(phi1, d1, spacing=s1)
        As1 = pi*phi1**2/4
        Asbr1 = As1 / s1
        d2 = 45.
        phi2 = 30.
        s2 = 100.
        as2 = SteelReinforcement.from_dimensions(phi2, d2, spacing=s2)
        As2 = pi*phi2**2/4
        Asbr2 = As2 / s2

        as12 = as1 + as2
        As12 = As1 + As2
        Asbr12 = Asbr1 + Asbr2
        self.assertAlmostEqual(as12.As, As12)
        self.assertAlmostEqual(as12.Asbr, Asbr12)
        d12 = (d1*Asbr1 + d2*Asbr2)/(Asbr1 + Asbr2)
        self.assertAlmostEqual(as12.d, d12)

        d3 = 42.
        phi3 = 30.
        s3 = 100.
        as3 = SteelReinforcement.from_dimensions(phi3, d3, spacing=s3)
        As3 = pi*phi3**2/4
        Asbr3 = As3 / s3

        as123 = as12 + as3
        As123 = As12 + As3
        Asbr123 = Asbr12 + Asbr3
        self.assertAlmostEqual(as123.As, As123)
        d123 = Asbr12*d12/Asbr123 + d3*Asbr3/Asbr123
        self.assertAlmostEqual(as123.d, d123)

        asother = as123 + 0.
        self.assertAlmostEqual(asother.As, As123)
        self.assertAlmostEqual(asother.Asbr, Asbr123)
        self.assertAlmostEqual(asother.d, d123)

    def test_raddition(self):
        d1 = 45.
        phi1 = 30.
        s1 = 100.
        as1 = SteelReinforcement.from_dimensions(phi1,  d1, spacing=s1)
        As1 = pi*phi1**2/4
        Asbr1 = As1 / s1

        asother = 0. + as1
        self.assertAlmostEqual(asother.As, As1)
        self.assertAlmostEqual(asother.Asbr, Asbr1)
        self.assertAlmostEqual(asother.d, d1)

    def test_phi(self):
        As = [314e-06, 380e-06, 28.3e-06, 804e-06]
        phi = [.020, .022, .006, .032]
        for i, Asi in enumerate(As):
            rebar = SteelReinforcement(As[i], 0.)
            with self.subTest(i=i):
                self.assertAlmostEqual(rebar.phi, phi[i])

        nbars = 4
        As = nbars * 380e-06
        rebar = SteelReinforcement(As, 0., nbars=nbars)
        self.assertEqual(rebar.phi, .022)


class TestPlanarElement(unittest.TestCase):

    def setUp(self):
        self.element = PlanarElement(thickness=10.)

    def test_thickness(self):
        self.assertAlmostEqual(self.element.thickness, 10.)

        self.element.thickness = 20.
        self.assertAlmostEqual(self.element.thickness, 20.)

    def test_material(self):
        self.assertIsNone(self.element.material)


class TestColumnDropPanel(unittest.TestCase):

    def setUp(self):
        self.column = RCColumn2D.create_rectangle_by_dimensions(1., 1.)
        self.lx = 0.5
        self.ly = 1.
        self.hH = 0.3
        self.drop_panel = ColumnDropPanel.create_as_column_offset(
                self.column, self.lx, self.hH, self.ly
                )

    def test_create_as_column_offset(self):
        self.assertIsInstance(self.drop_panel, ColumnDropPanel)
        self.assertEqual(self.drop_panel.centroid, self.column.centroid)
        self.assertAlmostEqual(self.drop_panel.height, self.hH)
        self.assertAlmostEqual(self.drop_panel.area, 6.)
        self.assertAlmostEqual(self.drop_panel.bx, 2.)
        self.assertAlmostEqual(self.drop_panel.by, 3.)

    def test_derived_properties(self):
        self.assertAlmostEqual(self.drop_panel.lx, self.lx)
        self.assertAlmostEqual(self.drop_panel.ly, self.ly)


class TestRCSlab(unittest.TestCase):

    def test_instantiation_by_factory(self):
        slab = RCSlab.create_rectangle_by_box(xymin=(0.0, 0.), xymax=(1., 1.),
                                              thickness=0.3)
        self.assertIsInstance(slab, RCSlab)
        self.assertAlmostEquals(slab.area, 1.)

    def test_add_agent(self):
        slab = RCSlab.create_rectangle_by_box(xymin=(0.0, 0.), xymax=(1., 1.),
                                              thickness=0.3)

        slab.add_agent(label='columns', agent_type=RCColumn2D, bx=0.3, by=0.3,
                       position='lower-left', origin=(0., 0.), thickness=0.4)
        slab.add_agent(label='columns', agent_type=RCColumn2D, bx=0.3, by=0.3,
                       position='lower-left', origin=(0., 0.), thickness=0.4)
        slab.add_agent(label='columns', agent_type=RCColumn2D, bx=0.3, by=0.3,
                       position='lower-left', origin=(0., 0.), thickness=0.4)

        self.assertEqual(len(slab.agents['columns']), 3)

        c = slab.agents['columns'][0]
        self.assertIsInstance(c, RCColumn2D)
        self.assertAlmostEquals(c.area, 0.09)

    def test_add_tension_bar_raw(self):
        slab = RCSlab.create_rectangle_by_box(xymin=(0.0, 0.), xymax=(1., 1.),
                                              thickness=0.3)
        # Set tension re-bar
        slab.add_tension_rebar_raw(phi=12., s=120., d=280.)
        slab.add_tension_rebar_raw(phi=16., s=120., d=295., axes='y', index=0)
        t_rebar = slab.t_rebar[0]
        self.assertAlmostEqual(t_rebar.value[0].As, pi*12.**2/4)
        self.assertAlmostEqual(t_rebar.value[1].As, pi*16.**2/4)
        self.assertAlmostEqual(t_rebar.value[0].Asbr, pi*12.**2/4/120.)
        self.assertAlmostEqual(t_rebar.value[1].Asbr, pi*16.**2/4/120.)
        self.assertAlmostEqual(t_rebar.value[0].d, 280.)
        self.assertAlmostEqual(t_rebar.value[1].d, 295.)

        # Replace tension re-bar
        slab.add_tension_rebar_raw(phi=16., s=120., d=295., axes='y', index=0,
                                   replace=True)
        self.assertAlmostEqual(t_rebar.value[0], 0.)
        self.assertAlmostEqual(t_rebar.value[1].As, pi*16.**2/4)
        self.assertAlmostEqual(t_rebar.value[1].Asbr, pi*16.**2/4/120.)
        self.assertAlmostEqual(t_rebar.value[1].d, 295.)

    def test_add_partial_rebar_raises(self):
        slab = RCSlab.create_rectangle_by_box(xymin=(0.0, 0.), xymax=(1., 1.),
                                              thickness=0.3)
        with self.assertRaises(ValueError):
            slab.add_partial_rebar(10., 100., 28., bx=10., by=1.)

    def test_add_partial_rebar(self):
        slab = RCSlab.create_rectangle_by_dimensions(
            bx=10., by=10, thickness=0.3, origin=(5., 5.)
            )
        slab.add_partial_rebar(10., 100., 28., bx=1., by=10., origin=(0.5, 5.))
        rebar = slab.rebar[0]
        self.assertAlmostEqual(rebar.geometry.area, 10.)

    def test_add_soil_pressure(self):
        slab = RCSlab.create_rectangle_by_box(xymin=(0., 0.), xymax=(1., 1.),
                                              thickness=0.)
        slab.add_soil_pressure(10.)
        self.assertAlmostEqual(slab.soil_pressure[0].value, 10.)



# Copyright (C) 2019 demetriou engineering ltd.
# Part of `dx-utilities`, licensed under the AGPLv3+
"""Base abstractions to handle safety-factors in
structural design codes.

We assume the following hierarchy:

* Design-situation

  * Limit-state, performance level etc.

    * Safety Factor
"""


class SafetyFactor(float):
    """A subclass of `float` with a couple of additional attributes.

    >>> from dx_base.safety_factors import SafetyFactor
    >>> sf = SafetyFactor(1.24, 'persistent', 'serviceability')
    >>> sf
    1.24
    >>> sf.design_situation
    'persistent'
    >>> sf.limit_stte
    'serviceability'

    :param float value: The value of the safety factor.
    :param str design_situation: Typically this is provided by
        the implemented design codes.
    :param str limit_state: This design limit-state.
    """

    def __new__(cls, value, *args, **kwargs):
        # Subclassing an immutable type requires setting
        # the __new__ method
        return super().__new__(cls, value)

    def __init__(self, value, design_situation='persistent',
                 limit_state='ultimate'):
        self.design_situation = design_situation
        self.limit_state = limit_state


class LimitStateFactor(SafetyFactor):
    """Safety factor for an arbitrary limit-state.

    :param str name: The name of state.
    :param family: The name of the family of states (if any).
    :param str design_situation:
    :type family: str or None
    """

    def __new__(cls, *args, value=1., **kwargs):
        return super().__new__(cls, value)

    def __init__(self, name, family=None, design_situation='persistent',
                 value=1.):
        self.name = name
        self.family = family
        super().__init__(value, design_situation=design_situation,
                         limit_state=str(self))

    def __str__(self):
        return '-'.join(filter(None, (self.name, self.family)))
